from django import forms
from django.utils.importlib import import_module
from avocado.conf import settings


def get_form_class(name):
    # Absolute import if a period exists, otherwise assume the
    # name refers to a built-in Django class
    if '.' in name:
        path = name.split('.')[:-1]
        module = import_module(path)
    else:
        if not name.endswith('Field'):
            name = name + 'Field'
        module = forms
    return getattr(module, name)


def get_internal_type(field):
    "Get model field internal type with 'field' off."
    datatype = field.get_internal_type().lower()
    if datatype.endswith('field'):
        datatype = datatype[:-5]
    return datatype


def get_heuristic_flags(field):
    # TODO add better conditions for determining how to set the
    # flags for most appropriate interface.
    # - Determine length of MAX value for string-based fields to rather
    # than relying on the `max_length`. This will enable checking TextFields
    # - Numerical fields may be enumerable, check the size of them if an
    # option is set?
    # For strings and booleans, set the enumerable flag by default
    # it below the enumerable threshold
    # TextFields are typically used for free text
    enumerable = False

    if field.internal_type != 'text' and field.simple_type in ('string', 'boolean') \
            and field.size <= settings.ENUMERABLE_MAXIMUM:
        enumerable = True

    return {
        'enumerable': enumerable,
    }



from django.template import Context
from django.template.loader import get_template
from _base import BaseExporter


class HTMLExporter(BaseExporter):
    short_name = 'HTML'
    long_name = 'HyperText Markup Language (HTML)'

    file_extension = 'html'
    content_type = 'text/html'

    preferred_formats = ('html', 'string')

    def write(self, iterable, template, buff=None, *args, **kwargs):
        buff = self.get_file_obj(buff)

        if isinstance(template, basestring):
            template = get_template(template)

        context = Context({'rows': self.read(iterable, *args, **kwargs)})
        buff.write(template.render(context))
        return buff



import warnings
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from modeltree.tree import trees
from modeltree.query import ModelTreeQuerySet
from django.core.exceptions import ValidationError


SORT_DIRECTIONS = ('asc', 'desc')


def has_keys(obj, keys):
    "Check the required keys are present in `obj`"
    for key in keys:
        if key not in obj:
            return False
    return True


class Node(object):
    def __init__(self, columns=None, ordering=None, **context):
        self.concept_ids = columns or []
        self.ordering = ordering or []
        self.tree = context.pop('tree', None)
        self.context = context

    def _get_concepts(self, ids):
        "Returns an ordered list of concepts based on `ids`."
        from avocado.models import DataConcept
        if not ids:
            return []
        concepts = list(DataConcept.objects.filter(pk__in=ids))
        concepts.sort(key=lambda o: ids.index(o.pk))
        return concepts

    def _get_fields_for_concepts(self, ids):
        "Returns an ordered list of fields for concept `ids`."
        from avocado.models import DataConceptField
        if not ids:
            return OrderedDict()
        # Concept fields that are sorted by concept then order, but are not
        # in the original order defined in `ids`
        cfields = list(DataConceptField.objects.filter(concept__pk__in=ids)\
            .select_related()\
            .order_by('concept', 'order'))

        # Order concept fields relative to `ids`
        cfields.sort(key=lambda o: ids.index(o.concept.pk))

        # Construct an ordered dict of fields by their concept
        groups = OrderedDict()
        for cf in cfields:
            pk = cf.concept.pk
            if pk not in groups:
                groups[pk] = []
            groups[pk].append(cf.field)
        return groups

    def _get_select(self, distinct):
        # Apply all fields to the query to ensure ordering get applied.
        # Django removes ORDER BY statements if column is not present in
        # SELECT since it will cause a SQL error. This ensures the ordering
        # is applied at the SQL level. The caveat here is that the rows
        # returned will include this extra data. The exporter classes handle
        # this by removing redundant rows relative to the *original* columns.
        ids = list(self.concept_ids)
        if self.ordering and distinct:
            ids += list(zip(*self.ordering)[0])

        # Flatten the grouped fields
        fields = [i for l in self._get_fields_for_concepts(ids).values() for i in l]

        model_fields = []

        for f in fields:
            if f.lexicon:
                model_fields.append(f.model._meta.get_field('label'))
            else:
                model_fields.append(f.field)
        return model_fields

    def _get_order_by(self):
        "Returns directional lookups to be unpacked in `QuerySet.order_by`."
        order_by = []

        if self.ordering:
            tree = trees[self.tree]
            ids, directions = zip(*self.ordering)
            groups = self._get_fields_for_concepts(ids)

            for pk, direction in self.ordering:
                for f in groups[pk]:
                    # Special case for Lexicon-based models, order by their
                    # model-defined `order` field.
                    if f.lexicon:
                        field = f.model._meta.get_field('order')
                        lookup = tree.query_string_for_field(field)
                    else:
                        lookup = tree.query_string_for_field(f.field)

                    if direction.lower() == 'desc':
                        order_by.append('-' + lookup)
                    else:
                        order_by.append(lookup)

        return order_by

    # Primary method for apply this view to a QuerySet
    def apply(self, queryset=None, include_pk=True):
        tree = trees[self.tree]
        if queryset is None:
            queryset = tree.get_queryset()

        # Convert to a ModelTreeQuerySet for the `select()` method
        queryset = ModelTreeQuerySet(tree, query=queryset.query)

        # Set model fields for `select()` method
        select = self._get_select(queryset.query.distinct)
        queryset = queryset.select(*select, include_pk=include_pk)

        # Set the order by on the QuerySet
        order_by = self._get_order_by()
        if order_by:
            queryset = queryset.order_by(*order_by)

        return queryset

    # Additional public methods for general use and interrogation
    def get_concepts_for_select(self):
        return self._get_concepts(self.concept_ids)

    def get_fields_for_select(self):
        return self._get_fields_for_concepts(self.concept_ids)

    def get_concepts_for_order_by(self):
        ids = []
        if self.ordering:
            ids = zip(*self.ordering)[0]
        return self._get_concepts(ids)

    def get_fields_for_order_by(self):
        ids = []
        if self.ordering:
            ids = zip(*self.ordering)[0]
        return self._get_fields_for_concepts(ids)

    @property
    def columns(self):
        warnings.warn('This property has been deprecated, used '
            'get_concepts_for_select() instead', DeprecationWarning)
        return self.get_concepts_for_select()


def validate(attrs, **context):
    if not attrs:
        return

    from avocado.models import DataConcept

    columns = attrs.get('columns', [])
    ordering = attrs.get('ordering', [])

    if columns:
        if len(set(columns)) != DataConcept.objects.filter(pk__in=columns).count():
            raise ValidationError('One or more concepts do not exist')

    if ordering:
        ids, directions = zip(*ordering)

        for _dir in directions:
            if _dir.lower() not in SORT_DIRECTIONS:
                raise ValidationError('Invalid sort direction')

        if len(set(ids)) != DataConcept.objects.filter(pk__in=ids).count():
            raise ValidationError('One or more concepts do not exist')


def parse(attrs, **context):
    if not attrs:
        return Node(**context)

    columns = attrs.get('columns', None)
    ordering = attrs.get('ordering', None)
    return Node(columns, ordering, **context)



import re
import os
import sys
import tempfile
import logging
from django.core import serializers, management
from django.db import (models, transaction, router, DEFAULT_DB_ALIAS,
        DatabaseError, IntegrityError)
from avocado.models import DataField, DataConcept, DataConceptField, DataCategory

FIXTURE_FORMAT = 'json'
FIXTURE_FILENAME_RE = re.compile(r'^[0-9]{4}_[0-9a-zA-Z_]+(\.json)$')

MIGRATION_MODEL_LABELS = ('avocado.datafield', 'avocado.dataconcept',
    'avocado.dataconceptfield', 'avocado.datacategory')

log = logging.getLogger(__name__)


def _fixture_filenames(dirname):
    filenames = []
    for f in os.listdir(dirname):
        if FIXTURE_FILENAME_RE.match(os.path.basename(f)):
            filenames.append(f)
    filenames.sort()
    return filenames


def next_fixture_name(name, dirname):
    filenames = _fixture_filenames(dirname)
    if not filenames:
        version = '0001'
    else:
        version = unicode(int(filenames[-1][:4]) + 1).zfill(4)
    return u'{0}_{1}'.format(version, name)


def full_fixture_path(name):
    return u'{0}.{1}'.format(os.path.join(get_fixture_dir(), name), FIXTURE_FORMAT)


def get_fixture_dir():
    from avocado.conf import settings
    fixture_dir = settings.METADATA_FIXTURE_DIR
    if fixture_dir:
        return fixture_dir
    app = models.get_app(settings.METADATA_MIGRATION_APP)
    if hasattr(app, '__path__'):
        # It's a 'models/' subpackage
        app_dir = os.path.dirname(os.path.dirname(app.__file__))
    else:
        # It's a models.py module
        app_dir = os.path.dirname(app.__file__)
    return os.path.join(app_dir, 'fixtures')


def create_fixture(name, using=DEFAULT_DB_ALIAS, silent=False):
    "Dumps data in the fixture format optionally to a particular path."
    if os.path.isabs(name):
        fixture_path = name
    else:
        fixture_path = full_fixture_path(name)
    with open(fixture_path, 'w') as fout:
        management.call_command('dumpdata', *MIGRATION_MODEL_LABELS,
            database=using, stdout=fout)
    if not silent:
        log.info(u'Created fixture {0}'.format(name))


def create_temp_fixture(*args, **kwargs):
    "Creates a fixture to a temporary file."
    fh, path = tempfile.mkstemp() 
    create_fixture(path, *args, **kwargs)
    return path


def load_fixture(name, using=DEFAULT_DB_ALIAS):
    """Progammatic way to load a fixture given some path. This does not
    assume the path is a fixture within some app and assumes a full path.
    """
    if os.path.isabs(name):
        fixture_path = name
    else:
        fixture_path = full_fixture_path(name)

    with open(fixture_path) as fixture:
        objects = serializers.deserialize(FIXTURE_FORMAT, fixture, using=using)

        with transaction.commit_manually(using):
            for obj in objects:
                if router.allow_syncdb(using, obj.object.__class__):
                    try:
                        obj.save(using=using)
                    except (DatabaseError, IntegrityError), e:
                        transaction.rollback(using)
                        msg = u'Could not load {app_label}.{object_name}(pk={pk}): {error_msg}'\
                            .format(app_label=obj.object._meta.app_label,
                                object_name=obj.object._meta.object_name,
                                pk=obj.object.pk, error_msg=e)
                        raise e.__class__, e.__class__(msg), sys.exc_info()[2]
            transaction.commit(using)
    log.info(u'Loaded data from fixture {0}'.format(name))


def delete_metadata(using=DEFAULT_DB_ALIAS):
    "Deletes all metadata in the target database."
    DataConceptField.objects.using(using).delete()
    DataConcept.objects.using(using).delete()
    DataField.objects.using(using).delete()
    DataCategory.objects.using(using).delete()
    log.debug(u'Metadata deleted from database "{0}"'.format(using))


def safe_load(name, backup_path=None, using=DEFAULT_DB_ALIAS):
    """Creates a backup of the current state of the metadata, attempts to load
    the new fixture and falls back to the backup fixture if the load fails for
    any reason.
    """
    with transaction.commit_manually(using):
        # Create the backup fixture 
        if backup_path:
            create_fixture(os.path.abspath(backup_path), using=using, silent=True)
        else:
            backup_path = create_temp_fixture(using=using, silent=True)
        log.info(u'Backup fixture written to {0}'.format(os.path.abspath(backup_path)))
        delete_metadata(using=using)
        try:
            load_fixture(name, using=using)
        except (DatabaseError, IntegrityError):
            transaction.rollback(using)
            log.error(u'Fixture load failed, reverting from backup: {0}'.format(backup_path))
            load_fixture(backup_path, using=using)
            raise
        transaction.commit(using)
    return backup_path



from django.utils import termcolors
from django import template

register = template.Library()


@register.simple_tag
def colorize(value, *opts, **kwargs):
    return termcolors.colorize(value, opts=opts, **kwargs) 



import logging
from django.db import models
from .managers import PublishedManager

log = logging.getLogger(__file__)


class Base(models.Model):
    """Base abstract class containing general metadata.

        ``name`` - The name _should_ be unique in practice, but is not enforced
        since in certain cases the name differs relative to the model and/or
        concepts these fields are asssociated with.

        ``description`` - Will tend to be exposed in client applications since
        it provides context to the end-users.

        ``keywords`` - Additional extraneous text that cannot be derived from
        the name, description or data itself. This is solely used for search
        indexing.
    """
    # Descriptor-based fields
    name = models.CharField(max_length=200, null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    keywords = models.CharField(max_length=100, null=True, blank=True)

    # Availability control mechanisms
    # Rather than deleting objects, they can be archived
    archived = models.BooleanField(default=False,
        help_text=u'Note: archived takes precedence over being published')
    # When `published` is false, it is globally not accessible.
    published = models.BooleanField(default=False)

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    objects = PublishedManager()

    class Meta(object):
        abstract = True

    def __unicode__(self):
        return unicode(self.name)

    @property
    def descriptors(self):
        return {
            'name': self.name,
            'description': self.description,
            'keywords': self.keywords,
        }

    def save(self, *args, **kwargs):
        if self.archived and self.published:
            self.published = False
            log.debug(u'{0} is published, but is being archived. It has ' \
                'been unpublished'.format(self))
        super(Base, self).save(*args, **kwargs)


class BasePlural(Base):
    """Adds field for specifying the plural form of the name.

        ``name_plural`` - Same as ``name``, but the plural form. If not
        provided, an 's' will appended to the end of the ``name``.
    """
    name_plural = models.CharField(max_length=200, null=True, blank=True)

    class Meta(object):
        abstract = True

    @property
    def descriptors(self):
        return {
            'name': self.name,
            'name_plural': self.name_plural,
            'description': self.description,
            'keywords': self.keywords,
        }

    def get_plural_name(self):
        if self.name_plural:
            plural = self.name_plural
        elif self.name and not self.name.endswith('s'):
            plural = self.name + 's'
        else:
            plural = self.name
        return plural



import sys
from django.db.models import Q
from django.core.management.base import BaseCommand
from avocado.models import DataField


class Command(BaseCommand):
    """
    SYNOPSIS::

        python manage.py avocado cache [options...] labels...

    DESCRIPTION:

        Finds all models referenced by the app, model or field `labels` and
        explicitly updates various cached properties relative to the
        `data_modified` on `DataField` instances.
    """

    help = '\n'.join([
        'Finds all models referenced by the app, model or field `labels` and',
        'explicitly updates various cached properties relative to the',
        '`data_modified` on `DataField` instances.',
    ])

    args = 'app [app.model, [app.model.field, [...]]]'

    def _progress(self):
        sys.stdout.write('.')
        sys.stdout.flush()

    def handle(self, *args, **options):
        "Handles app_label or app_label.model_label formats."
        conditions = []

        for label in args:
            labels = label.split('.')

            # Specific field
            if len(labels) == 3:
                app, model, field = labels
                conditions.append(Q(app_name=app, model_name=model, field_name=field))
            # All fields for a model
            elif len(labels) == 2:
                app, model = labels
                conditions.append(Q(app_name=app, model_name=model))
            # All fields for each model in the app
            else:
                app = labels[0]
                conditions.append(Q(app_name=app))

        fields = DataField.objects.filter(enumerable=True)
        if conditions:
            q = Q()
            for x in conditions:
                q = q | x
            fields = fields.filter(q).distinct()

        count = 0
        for datafield in fields:
            datafield.size
            self._progress()
            datafield.labels
            self._progress()
            datafield.values
            self._progress()
            if datafield.lexicon:
                datafield.coded
                self._progress()
            count += 1

        print(u'{0} DataFields have been updated'.format(count))



from zipfile import ZipFile
from cStringIO import StringIO
from string import punctuation
from django.template import Context
from django.template.loader import get_template
from _base import BaseExporter
from _csv import CSVExporter


class RExporter(BaseExporter):
    short_name = 'R'
    long_name = 'R Programming Language'

    file_extension = 'zip'
    content_type = 'application/zip'

    preferred_formats = ('r', 'coded', 'number', 'string')

    def _format_name(self, name):
        punc = punctuation.replace('_', '')
        name = str(name).translate(None, punc)
        name = name.replace(' ', '_')
        words = name.split('_')
        for i, w in enumerate(words):
            if i == 0:
                name = w.lower()
                continue
            name += w.capitalize()
        if name[0].isdigit():
            name = '_' + name

        return name

    def _code_values(self, name, field):
        "If the field can be coded return the r factor and level for it."
        data_field = u'data${0}'.format(name)

        factor = u'{0}.factor = factor({0},levels=c('.format(data_field)
        level = u'levels({0}.factor)=c('.format(data_field)

        values_len = len(field.codes)

        for i, (code, label) in enumerate(field.coded_choices):
            factor += unicode(code)
            level += u'"{0}"'.format(unicode(label))
            if i == values_len - 1:
                factor += '))\n'
                level += ')\n'
                continue
            factor += ' ,'
            level += ' ,'
        return factor, level

    def write(self, iterable, buff=None, template_name='export/script.R', *args, **kwargs):
        zip_file = ZipFile(self.get_file_obj(buff), 'w')

        factors = []      # field names
        levels = []       # value dictionaries
        labels = []       # data labels

        for c in self.concepts:
            cfields = c.concept_fields.all()
            for cfield in cfields:
                field = cfield.field
                name = self._format_name(field.field_name)
                labels.append(u'attr(data${0}, "label") = "{1}"'.format(name, unicode(cfield)))

                if field.lexicon:
                    codes = self._code_values(name, field)
                    factors.append(codes[0])
                    levels.append(codes[1])

        data_filename = 'data.csv'
        script_filename = 'script.R'

        # File buffers
        data_buff = StringIO()
        # Create the data file
        data_exporter = CSVExporter(self.concepts)
        # Overwrite preferred formats for data file
        data_exporter.preferred_formats = self.preferred_formats
        data_exporter.write(iterable, data_buff, *args, **kwargs)

        zip_file.writestr(data_filename, data_buff.getvalue())

        template = get_template(template_name)
        context = Context({
            'data_filename': data_filename,
            'labels': labels,
            'factors': factors,
            'levels': levels,
        })

        # Write script from template
        zip_file.writestr(script_filename, template.render(context))
        zip_file.close()

        return zip_file



# A mapping between model field internal datatypes and sensible
# client-friendly datatypes. In virtually all cases, client programs
# only need to differentiate between high-level types like number, string,
# and boolean. More granular separation be may desired to alter the
# allowed operators or may infer a different client-side representation
SIMPLE_TYPE_MAP = {
    'auto': 'key',
    'foreignkey': 'key',

    'biginteger': 'number',
    'decimal': 'number',
    'float': 'number',
    'integer': 'number',
    'positiveinteger': 'number',
    'positivesmallinteger': 'number',
    'smallinteger': 'number',

    'nullboolean': 'boolean',

    'char': 'string',
    'email': 'string',
    'file': 'string',
    'filepath': 'string',
    'image': 'string',
    'ipaddress': 'string',
    'slug': 'string',
    'text': 'string',
    'url': 'string',
}

# A mapping between the client-friendly datatypes and sensible operators
# that will be used to validate a query condition. In many cases, these types
# support more operators than what are defined, but are not include because
# they are not commonly used.
OPERATOR_MAP = {
    'key': ('exact', '-exact', 'in', '-in'),
    'boolean': ('exact', '-exact', 'in', '-in'),
    'date': ('exact', '-exact', 'in', '-in', 'lt', 'lte', 'gt', 'gte', 'range'),
    'number': ('exact', '-exact', 'in', '-in', 'lt', 'lte', 'gt', 'gte', 'range'),
    'string': ('exact', '-exact', 'iexact', '-iexact', 'in', '-in', 'icontains', '-icontains'),
    'datetime': ('exact', '-exact', 'in', '-in', 'lt', 'lte', 'gt', 'gte', 'range'),
    'time': ('exact', '-exact', 'in', '-in', 'lt', 'lte', 'gt', 'gte', 'range'),
}

# A general mapping of formfield overrides for all subclasses. the mapping is
# similar to the SIMPLE_TYPE_MAP, but the values reference internal
# formfield classes, that is integer -> IntegerField. in many cases, the
# validation performed may need to be a bit less restrictive than what the
# is actually necessary
INTERNAL_DATATYPE_FORMFIELDS = {
    'integer': 'FloatField',
    'positiveinteger': 'FloatField',
    'positivesmallinteger': 'FloatField',
    'smallinteger': 'FloatField',

    # Generic datatypes mapped from above
    'number': 'FloatField',
}

# The minimum number of distinct values required when determining to set the
# `searchable` flag on `DataField` instances during the `init` process. This
# will only be applied to fields with a Avocado datatype of 'string'
ENUMERABLE_MAXIMUM = 30

# Flag for enabling the history API
HISTORY_ENABLED = False

# The maximum size of a user's history. If the value is an integer, this
# is the maximum number of allowed items in the user's history. Set to
# `None` (or 0) to enable unlimited history. Note, in order to enforce this
# limit, the `avocado history --prune` command must be executed to remove
# the oldest history from each user based on this value.
HISTORY_MAX_SIZE = None

# App that the metadata migrations will be created for. This is typically the
# project itself.
METADATA_MIGRATION_APP = None

# Directory for the migration backup fixtures. If None, this will default to
# the fixtures dir in the app defined by `METADATA_MIGRATION_APP`
METADATA_FIXTURE_DIR = None

METADATA_FIXTURE_SUFFIX = 'avocado_metadata'

METADATA_MIGRATION_SUFFIX = 'avocado_metadata_migration'
